const crypto = require('crypto')
const dh1024 = require('../utils/dh1024')
const {sha1} = require('../utils/crypto')

const DH_P = Buffer.from([
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xc9, 0x0f, 0xda, 0xa2, 0x21, 0x68, 0xc2, 0x34,
  0xc4, 0xc6, 0x62, 0x8b, 0x80, 0xdc, 0x1c, 0xd1, 0x29, 0x02, 0x4e, 0x08, 0x8a, 0x67, 0xcc, 0x74,
  0x02, 0x0b, 0xbe, 0xa6, 0x3b, 0x13, 0x9b, 0x22, 0x51, 0x4a, 0x08, 0x79, 0x8e, 0x34, 0x04, 0xdd,
  0xef, 0x95, 0x19, 0xb3, 0xcd, 0x3a, 0x43, 0x1b, 0x30, 0x2b, 0x0a, 0x6d, 0xf2, 0x5f, 0x14, 0x37,
  0x4f, 0xe1, 0x35, 0x6d, 0x6d, 0x51, 0xc2, 0x45, 0xe4, 0x85, 0xb5, 0x76, 0x62, 0x5e, 0x7e, 0xc6,
  0xf4, 0x4c, 0x42, 0xe9, 0xa6, 0x37, 0xed, 0x6b, 0x0b, 0xff, 0x5c, 0xb6, 0xf4, 0x06, 0xb7, 0xed,
  0xee, 0x38, 0x6b, 0xfb, 0x5a, 0x89, 0x9f, 0xa5, 0xae, 0x9f, 0x24, 0x11, 0x7c, 0x4b, 0x1f, 0xe6,
  0x49, 0x28, 0x66, 0x51, 0xec, 0xe6, 0x53, 0x81, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
]);

class KeyAgreementTap {
  constructor (or) {
    this._router = or;
    this._dh = crypto.createDiffieHellman(DH_P);
    this._dh.generateKeys();
  }

  get public_key() {
    return this._dh.getPublicKey();
  }

  get private_key() {
    return this._dh.getPrivateKey();
  }

  compute_shared_secret(handshake_data) {
    return this._compute_shared_secret(
      handshake_data.slice(0, dh1024.key_size_in_bytes),
      handshake_data.slice(dh1024.key_size_in_bytes, dh1024.key_size_in_bytes + 20));
  }

  _compute_shared_secret(other_public_key, verification_data) {
    //
    // 5.1.3. The "TAP" handshake
    //
    // This handshake uses Diffie-Hellman in Z_p and RSA to compute a set of
    // shared keys which the client knows are shared only with a particular
    // server, and the server knows are shared with whomever sent the
    // original handshake (or with nobody at all).  It's not very fast and
    // not very good.  (See Goldberg's "On the Security of the Tor
    // Authentication Protocol".)
    //
    // Define TAP_C_HANDSHAKE_LEN as DH_LEN+KEY_LEN+PK_PAD_LEN.
    // Define TAP_S_HANDSHAKE_LEN as DH_LEN+HASH_LEN.
    //
    // The payload for a CREATE cell is an 'onion skin', which consists of
    // the first step of the DH handshake data (also known as g^x).  This
    // value is hybrid-encrypted (see 0.3) to the server's onion key, giving
    // a client handshake of:
    //
    //     PK-encrypted:
    //       Padding                       [PK_PAD_LEN bytes]
    //       Symmetric key                 [KEY_LEN bytes]
    //       First part of g^x             [PK_ENC_LEN-PK_PAD_LEN-KEY_LEN bytes]
    //     Symmetrically encrypted:
    //       Second part of g^x            [DH_LEN-(PK_ENC_LEN-PK_PAD_LEN-KEY_LEN)
    //                                         bytes]
    //
    // The payload for a CREATED cell, or the relay payload for an
    // EXTENDED cell, contains:
    //       DH data (g^y)                 [DH_LEN bytes]
    //       Derivative key data (KH)      [HASH_LEN bytes]   <see 5.2 below>
    //
    // Once the handshake between the OP and an OR is completed, both can
    // now calculate g^xy with ordinary DH.  Before computing g^xy, both parties
    // MUST verify that the received g^x or g^y value is not degenerate;
    // that is, it must be strictly greater than 1 and strictly less than p-1
    // where p is the DH modulus.  Implementations MUST NOT complete a handshake
    // with degenerate keys.  Implementations MUST NOT discard other "weak"
    // g^x values.
    //
    // (Discarding degenerate keys is critical for security; if bad keys
    // are not discarded, an attacker can substitute the OR's CREATED
    // cell's g^y with 0 or 1, thus creating a known g^xy and impersonating
    // the OR. Discarding other keys may allow attacks to learn bits of
    // the private key.)
    //
    // Once both parties have g^xy, they derive their shared circuit keys
    // and 'derivative key data' value via the KDF-TOR function in 5.2.1.
    //

    //mini_assert(verification_data.get_size() == crypto::sha1::hash_size_in_bytes);

    const shared_secret = this._dh.computeSecret(other_public_key);
    const derived = this._derive_keys(shared_secret);

    //
    // first 20 bytes of the derived key is the verification checksum.
    // rest of it is the key material.
    //
    const computed_verification_data = derived.slice(0, 20);
    const key_material               = derived.slice(20);

    if (computed_verification_data.equals(verification_data))
    {
      return key_material;
    }

    return Buffer.alloc(0);
  }

  _derive_keys(secret) {
    const key_material = Buffer.alloc(100);

    const hashdata = Buffer.alloc(secret.length + 1);
    secret.copy(hashdata, 0);

    for (let i = 0; i < 5; i++)
    {
      hashdata[secret.length] = i;
      sha1(hashdata).copy(key_material, i * 20);
    }

    return key_material;
  }

}
module.exports = KeyAgreementTap;