const crypto = require('crypto');
/**
 * @returns {Buffer}
 */
function hmac_sha256(key, data) {
  const h = crypto.createHmac('sha256', key);
  return h.update(data).digest();
}
exports.hmac_sha256 = hmac_sha256;

/**
 * @param {BinaryLike} data
 * @returns {Buffer}
 */
function sha1(data) {
  const h = crypto.createHash('sha1');
  return h.update(data).digest();
}
exports.sha1 = sha1;

function rsa_1024(public_key) {
  const rsa = crypto.createPublicKey(public_key);
  return {
    encrypt(data) {
      return crypto.publicEncrypt(rsa, data);
    }
  }
}

function aes_ctr_128(key) {
  const aes = crypto.createCipheriv('aes-128-ctr', key, Buffer.alloc(16));
  return {
    encrypt(data) {
      return aes.update(data);
    }
  }
}

const KEY_LEN = 16;
const PK_ENC_LEN = 128;
const PK_PAD_LEN = 42;
const PK_DATA_LEN = PK_ENC_LEN - PK_PAD_LEN;
const PK_DATA_LEN_WITH_KEY = PK_DATA_LEN - KEY_LEN;

/**
 * @param {Buffer} data
 * @param {string} public_key
 * @returns {Buffer}
 */
function hybrid_encrypt(data, public_key) {
  if (data.length < PK_DATA_LEN)
  {
    return crypto.publicEncrypt(public_key, data);
  }

  const random_key = crypto.randomBytes(KEY_LEN);

  //
  // RSA( K | M1 ) --> C1
  //
  const k_and_m1 = Buffer.concat([
    random_key, data.slice(0, PK_DATA_LEN_WITH_KEY)
  ], random_key.length + PK_DATA_LEN_WITH_KEY);

  const c1 = crypto.publicEncrypt(public_key, k_and_m1);

  //
  // AES_CTR(M2)  --> C2
  //
  const m2 = data.slice(PK_DATA_LEN_WITH_KEY);
  const c2 = aes_ctr_128(random_key).encrypt(m2);

  //
  // C1 | C2
  //
  return  Buffer.concat([ c1, c2 ], c1.length + c2.length);
}
exports.hybrid_encrypt = hybrid_encrypt;
